from manhattan.chains import Chain
from unittest.mock import MagicMock


def test_init():
    """Initialize a chain"""
    chain = Chain(['foo', 'bar'])
    assert chain.links == ['foo', 'bar']

def test_call():
    """Call the chain executing the linked functions"""
    chain = Chain([
        'foo',
        'bar',
        [
            ['zee'],
            ['omm']
        ]
    ])

    @chain.link
    def foo(state):
        pass

    @chain.link
    def bar(state):
        if state.bar:
            return 'bar'
        return state.zee

    @chain.link
    def zee(state):
        return 'zee'

    @chain.link
    def omm(state):
        return 'omm'

    # Check the chain exectution works as expected
    assert chain(bar=False, zee=False) == 'omm'
    assert chain(bar=False, zee=True) == 'zee'
    assert chain(bar=True, zee=True) == 'bar'

def test_links():
    """Get a copy of the links in the chain"""
    links = [
        'foo',
        'bar',
        [
            ['zee'],
            ['omm']
        ]
    ]
    chain = Chain(links)

    # Check the links are the same
    assert chain.links == links

    # Check the links are a copy (deepcopy)
    assert chain.links is not links
    assert chain.links[2] is not links[2]
    assert chain.links[2][0] is not links[2][0]
    assert chain.links[2][1] is not links[2][1]

def test_copy():
    """Copy the chain"""
    chain = Chain(['foo', 'bar'])

    @chain.link
    def foo(state):
        pass

    @chain.link
    def bar(state):
        pass

    chain_copy = chain.copy()

    # Check the chain is the same
    assert chain_copy.links == chain.links
    assert chain_copy._funcs == chain._funcs

    # Check the chain is a copy
    assert chain_copy is not chain
    assert chain_copy._links is not chain._links
    assert chain_copy._funcs is not chain._funcs

def test_get_final():
    """Get the function set as the final function for the chain"""

    chain = Chain(['foo', 'bar'])

    def zee(state):
        pass

    chain.set_final(zee)

    assert zee == chain.get_final()

def test_get_link():
    """Get the function mapped to the named link(s) in the chain"""
    chain = Chain(['foo', 'bar'])

    @chain.link
    def foo(state):
        pass

    assert chain.get_link('foo') == foo

def test_insert_link():
    """Insert a new link into the chain"""

    chain = Chain([
        'foo',
        [
            ['bar'],
            ['zee']
        ]
    ])

    # Check we can insert a link in the root route
    chain.insert_link('foo', 'omm')
    assert chain.links == [
        'omm',
        'foo',
        [
            ['bar'],
            ['zee']
        ]
    ]

    # Check we can insert a link in an intersection route
    chain.insert_link('bar', 'yaa')
    assert chain.links == [
        'omm',
        'foo',
        [
            ['yaa', 'bar'],
            ['zee']
        ]
    ]

    # Check we can insert a link in after the target link
    chain.insert_link('zee', 'huu', True)
    assert chain.links == [
        'omm',
        'foo',
        [
            ['yaa', 'bar'],
            ['zee', 'huu']
        ]
    ]

def test_remove_link():
    """Remove a link from the chain"""

    chain = Chain([
        'foo',
        'bar',
        [
            ['zee', 'omm'],
            ['yaa']
        ]
    ])

    # Check we can remove from the root route
    chain.remove_link('bar')
    assert chain.links == [
        'foo',
        [
            ['zee', 'omm'],
            ['yaa']
        ]
    ]

    # Check we can remove from an intersection route
    chain.remove_link('zee')
    assert chain.links == [
        'foo',
        [
            ['omm'],
            ['yaa']
        ]
    ]

def test_set_final():
    """Set the final function for a chain"""
    chain = Chain(['foo', 'bar'])

    # Set links
    @chain.link
    def foo(state):
        pass

    @chain.link
    def bar(state):
        pass

    # Set the final function
    zee = MagicMock()
    chain.set_final(zee)

    # Call the chain
    chain()

    assert zee.called

def test_set_links():
    """Map a function to the named link(s) in the chain"""
    chain = Chain(['foo', 'bar'])

    # Set the link using the functions name
    def foo(state):
        pass

    chain.set_link(foo)
    assert chain.get_link('foo') == foo

    # Set the link using a specified name
    def zee(state):
        pass

    chain.set_link(zee, name='bar')
    assert chain.get_link('bar') == zee

def test_final_decorator():
    """Set the final function for a chain"""
    chain = Chain(['foo', 'bar'])

    # Set links
    @chain.link
    def foo(state):
        pass

    @chain.link
    def bar(state):
        pass

    # Set the final function
    zee_mock = MagicMock()

    @chain.final
    def zee(state):
        zee_mock()

    chain()

    assert zee_mock.called

def test_link_decorator():
    """Map a function to the named link(s) in the chain"""
    chain = Chain(['foo', 'bar'])

    # Set the link using the functions name
    @chain.link
    def foo(state):
        pass

    assert chain.get_link('foo') == foo

    # Set the link using a specified name
    @chain.link(name='bar')
    def zee(state):
        pass

    assert chain.get_link('bar') == zee

def test_update():
    """Update the links for the chain"""
    chain = Chain(['foo', 'bar'])

    @chain.link
    def foo(state):
        pass

    @chain.link
    def bar(state):
        pass

    chain.update(['foo', 'zee'])

    # Check the link structure has been updated
    assert chain.links == ['foo', 'zee']



from flask import Flask

from manhattan.chains import Chain, ChainMgr


def test_init():
    """Initialize a chain manager"""
    chains = ChainMgr()

    # Initially the chain manager should have no assigned chains
    assert chains.chains == {}

def test_assign_chain():
    """Assign a chain to the chain manager"""
    chains = ChainMgr()

    my_chain = Chain(['foo', 'bar'])
    chains['my_chain'] = my_chain

    assert chains['my_chain'] is my_chain

def test_chains_property():
    """Get a map of chains for the chain manager using the chains property"""
    chains = ChainMgr()

    my_chain_a = Chain(['foo', 'bar'])
    my_chain_b = Chain(['zee', 'omm'])
    chains['my_chain_a'] = my_chain_a
    chains['my_chain_b'] = my_chain_b

    assert chains.chains == {'my_chain_a': my_chain_a, 'my_chain_b': my_chain_b}

def test_copy():
    """Copy the chain manager"""
    chains = ChainMgr()

    my_chain_a = Chain(['foo', 'bar'])
    my_chain_b = Chain(['zee', 'omm'])
    chains['my_chain_a'] = my_chain_a
    chains['my_chain_b'] = my_chain_b

    chains_copy = chains.copy()

    # Check the chain manager is the same
    assert chains.chains == chains.chains

    # Check the chain is a copy
    assert chains_copy is not chains
    assert chains_copy._chains is not chains._chains

def test_retrieve_chain():
    """Retreive a chain from the chain manager by name"""
    chains = ChainMgr()

    my_chain = Chain(['foo', 'bar'])
    chains['my_chain'] = my_chain

    assert chains['my_chain'] is my_chain

def test_delete_chain():
    """Delete a chain from the chain manager by name"""
    chains = ChainMgr()

    my_chain = Chain(['foo', 'bar'])
    chains['my_chain'] = my_chain
    del chains['my_chain']

    assert 'my_chain' not in chains

def test_insert_link():
    """Insert a link into every chain"""

    chains = ChainMgr()

    my_chain_a = Chain(['foo', 'bar'])
    my_chain_b = Chain(['zee', 'bar'])
    chains['my_chain_a'] = my_chain_a
    chains['my_chain_b'] = my_chain_b

    chains.insert_link('bar', 'omm', after=True)

    assert chains['my_chain_a'].links == ['foo', 'bar', 'omm']
    assert chains['my_chain_b'].links == ['zee', 'bar', 'omm']

def test_remove_link():
    """Remove a link from every chain"""

    chains = ChainMgr()

    my_chain_a = Chain(['foo', 'bar'])
    my_chain_b = Chain(['zee', 'bar'])
    chains['my_chain_a'] = my_chain_a
    chains['my_chain_b'] = my_chain_b

    chains.remove_link('bar')

    assert chains['my_chain_a'].links == ['foo']
    assert chains['my_chain_b'].links == ['zee']

def test_set_final():
    """Set a final function across one or more chains in the chain manager"""
    chains = ChainMgr()

    chains['my_chain_a'] = Chain(['foo', 'bar'])
    chains['my_chain_b'] = Chain(['foo', 'bar'])

    # Add a final function to just one chain
    def zee(state):
        pass

    chains.set_final(zee, scope={'my_chain_a'})

    assert chains['my_chain_a'].get_final() is zee
    assert chains['my_chain_b'].get_final() is None

    # Add a link to multiple chains
    def omm(state):
        pass

    chains.set_final(omm)

    assert chains['my_chain_a'].get_final() is omm
    assert chains['my_chain_b'].get_final() is omm

def test_set_link():
    """Set a link across one or more chains in the chain manager"""
    chains = ChainMgr()

    chains['my_chain_a'] = Chain(['foo', 'bar'])
    chains['my_chain_b'] = Chain(['foo', 'bar'])

    # Add a link to just one chain
    def foo(state):
        pass

    chains.set_link(foo, scope={'my_chain_a'})

    assert chains['my_chain_a'].get_link('foo') is foo
    assert chains['my_chain_b'].get_link('foo') is None

    # Add a final function to multiple chains
    def bar(state):
        pass

    chains.set_link(bar)

    assert chains['my_chain_a'].get_link('bar') is bar
    assert chains['my_chain_b'].get_link('bar') is bar

def test_final_decorator():
    """Set a final function across one or more chains in the chain manager"""
    chains = ChainMgr()

    chains['my_chain_a'] = Chain(['foo', 'bar'])
    chains['my_chain_b'] = Chain(['foo', 'bar'])

    # Add a final function to just one chain
    @chains.final(scope={'my_chain_a'})
    def zee(state):
        pass

    assert chains['my_chain_a'].get_final() is zee
    assert chains['my_chain_b'].get_final() is None

    # Add a final function to multiple chains
    @chains.final
    def omm(state):
        pass

    assert chains['my_chain_a'].get_final() is omm
    assert chains['my_chain_b'].get_final() is omm

def test_link_decorator():
    """
    Set a link across one or more chains in the chain manager with the link
    decorator.
    """
    chains = ChainMgr()

    chains['my_chain_a'] = Chain(['foo', 'bar'])
    chains['my_chain_b'] = Chain(['foo', 'bar'])

    # Add a link to just one chain
    @chains.link(scope={'my_chain_a'})
    def foo(state):
        pass

    assert chains['my_chain_a'].get_link('foo') is foo
    assert chains['my_chain_b'].get_link('foo') is None

    # Add a link to multiple chains
    @chains.link
    def bar(state):
        pass

    assert chains['my_chain_a'].get_link('bar') is bar
    assert chains['my_chain_b'].get_link('bar') is bar

def test_flask_view():
    """
    Return a Flask view that will call a chain against the manager with the
    appropriate HTTP method.
    """
    chains = ChainMgr()

    chains['get'] = Chain(['getter'])
    chains['post'] = Chain(['poster'])

    @chains.link(scope={'get'})
    def getter(state):
        return 'get'

    @chains.link(scope={'post'})
    def poster(state):
        return 'post'

    # Create a test application to run
    app = Flask(__name__)
    app.add_url_rule(
        '/',
        'index',
        chains.flask_view(zee='omm'),
        methods=['GET', 'POST']
        )
    client = app.test_client()

    # Test get request
    res = client.get('/')
    assert res.data == b'get'

    # Test post request
    res = client.post('/')
    assert res.data == b'post'



from manhattan.chains import State


def test_state_from_dict():
    """Initialize a state from a dictionary"""
    state = State({'foo': 'bar'})
    assert state == {'foo': 'bar'}

def test_state_from_kwargs():
    """Initalize a state from a set of keyword arguments"""
    state = State(foo='bar')
    assert state == {'foo': 'bar'}

def test_state_assignment():
    """Assign an item to the state using dot-notation"""
    state = State()
    state.foo = 'bar'
    assert state == {'foo': 'bar'}

def test_state_retrieval():
    """Retrieve an item from the state using dot-notation"""
    state = State({'foo': 'bar'})
    assert state.foo == 'bar'

def test_state_deletion():
    """Delete an item from the state using dot-notation"""
    state = State({'foo': 'bar'})
    del state.foo
    assert 'foo' not in state


"""
Classes for implementing execution chains.
"""

from copy import deepcopy
import json
import pprint
import sys
import time

try:
    # The Flask `request` module is required for the `flask_view` method, but we
    # treat this as an option requirement since chains this may not be
    # applicable when using chains.
    from flask import request as flask_request
except ImportError:
    pass


class ChainMgr:
    """
    A class for managing multiple related chains, e.g:

        chains = ChainMgr()
        chains['get']  = Chain(...)
        chains['post'] = Chain(...)

        method = request.method.lower()
        if method in chains:
            chains[method]()
        else:
            abort(405)
    """

    def __init__(self):
        # A map of named chains
        self._chains = {}

    def __contains__(self, name):
        return name in self._chains

    def __getitem__(self, name):
        return self._chains[name]

    def __setitem__(self, name, value):
        self._chains[name] = value

    def __delitem__(self, name):
        del self._chains[name]

    @property
    def chains(self):
        """Return a copy of the chains map for the manager"""
        return self._chains.copy()

    def copy(self):
        """Return a copy of the chain manager"""
        chains = self.__class__()
        for name, chain in self._chains.items():
            chains[name] = chain.copy()
        return chains

    def insert_link(self, target_link, new_link, after=False):
        """
        Insert a new link in every chain before/after every instance of the
        target link. By default new links are inserted before the target link,
        to insert the new link after the target link pass `after` as True to the
        method.
        """
        for chain in self._chains.values():
            chain.insert_link(target_link, new_link, after=after)

    def remove_link(self, target_link):
        """Remove all instances of a link from every chain"""
        for chain in self._chains.values():
            chain.remove_link(target_link)

    def set_final(self, func, scope=None):
        """Sets the final function across one or more chains"""

        if scope is None:
            # If no scope is specified then add the link to all chains
            scope = self._chains.keys()

        for name in scope:
            self[name].set_final(func)

    def set_link(self, func, scope=None, **kwargs):
        """
        Sets a function for a link. Works exactly like the `link()` decorator,
        e.g:

            # Set for all chains
            def foo(state):
                ...

            chains.set_link(foo)

            # Set for the given scope of chains
            def foo(state):
                ...

            chains.add_link(foo, scope={'get', 'post'})

            # Set for the link named 'bar'
            def foo(state):
                ...

            chains.add_link(foo, scope={'get', 'post'}, name='bar')
        """

        if scope is None:
            # If no scope is specified then add the link to all chains
            scope = self._chains.keys()

        for name in scope:
            self[name].set_link(func, **kwargs)

    def super(self, state, chain_name=None):
        """
        Call super against the relevant chain based on the current flask
        request.
        """

        if not chain_name and 'flask' in sys.modules:

            # If not chain name is given, default to the flask request method
            # as this is most commonly what is required.
            chain_name = flask_request.method.lower()

        return self[chain_name].super(state)

    # Decorators

    def final(self, *args, **kwargs):
        """Sets the final function across one or more chains"""

        def wrap(func):
            if 'scope' in kwargs:
                scope = kwargs.pop('scope')
            else:
                # If no scope is specified then set the final for all chains
                scope = self._chains.keys()

            for name in scope:
                self._chains[name].set_final(func)

            return func

        if len(args) == 1 and callable(args[0]):
            return wrap(args[0])
        else:
            return wrap

    def link(self, *args, **kwargs):
        """
        A decorator that will set the decorated function as a link in the
        given set of chains, e.g:

            # Set for all chains
            @chains.link
            def foo(state):
                ...

            # Set for the given scope of chains
            @chains.link(scope={'get', 'post'})
            def foo(state):
                ...

            # Set for the link named 'bar'
            @chains.link(scope={'get', 'post'}, name='bar')
            def foo(state):
                ...
        """
        def wrap(func):
            if 'scope' in kwargs:
                scope = kwargs.pop('scope')
            else:
                # If no scope is specified then add the link to all chains
                scope = self._chains.keys()

            for name in scope:
                self._chains[name].set_link(func, **kwargs)

            return func

        if len(args) == 1 and callable(args[0]):
            return wrap(args[0])
        else:
            return wrap

    # Framework methods

    def flask_view(self, **kwargs):
        """
        Return a Flask view function for the chain. The keyword arguments are
        used to initialize the chain when called. The request.method (lowercase)
        value is used to determine which chain get's called, therefore the chain
        manager must support at least one applicable HTTP verb.
        """

        def _view():
            assert 'flask' in sys.modules, \
                'Flask must be installed to use this method'

            method = flask_request.method.lower()
            if method in self:
                return self[method](**kwargs)

        return _view


class Chain:
    """
    Chains provide a pattern for *loosely* coupling together a series of
    functions with support for intersections.

    Within a chain logic is broken down into small independent blocks (known
    as links), these links can be easily switched out and/or re-arranged to
    modify the behaviour of the chain.

    Chains are initially defined as a list of named links without associated
    function, for example:

        # An puesdo chain for updating a document via a HTTP POST request
        chain = Chain([
            'lookup',
            'check_permissions',
            'init_form',
            'validate',
            [
                [
                    'update',
                    'redirect'
                ],
                ['template']
            ]
        ])

    The function for each link can then be added using the `set_link` method or
    more typically the `link` decorator like so:

        @chain.link
        def lookup(state):
            state.product = Product.by_id(request.values.get('id'))
            if not state.product:
                abort(404)

    The flow of execution is controlled by the return value of the link, and
    these rules are simple but important:

    - If `None` is returned then call the next link.
    - If a `True` or `False` is returned then this determines the fork taken at
      an intersection.
    - If any other value is returned then the execution is stoped and the value
      returned to the caller.

    An intersection is represented by a list containing two lists of named
    links, as in the example chain above where an intersection follows the
    'validate' link, validate might look something like this:

        @chain.link
        def validate(state):
            return scope.form.validate()

    If the form is valid then 'update' will be exectuted next, otherwise
    'template' will be, at which template will return a response in form of a
    render template and execution will end (even if there are other links after
    this):

        @chain.link
        def template(state):
            # We send the entire scope to the template as this makes it easier
            return render_template('products/update.html', **scope)

    By breaking the logic down in this way chains make it easy to modify their
    behaviour from anywhere. The initial reason for designing the chain pattern
    was to allow controllers for models in our web applications (which provide
    standard functions for add, update, delete, list, view, etc) to be easily
    modified without the need to override methods which typically contain
    complete chunks of logic (such as everything in our example chain).
    """

    def __init__(self, links):
        # A structured list of links that make up the chain
        self._links = links

        # A map of links to functions
        self._funcs = {}

        # A function guaranteed to be called after the chain is executed
        self._final = None

    def __call__(self, _trace=False, **kwargs):
        """Call the chain of links"""

        # Create a state object that will be passed to (and potentially
        # mutated by) each link in the chain.
        state = State(**kwargs)
        state._call = State(
            route=[],
            trace=[]
        )

        # We call the chain within a try..finally block
        try:
            r = self._call(_trace, state)

            if _trace:

                # If the call was traced print details of the trace
                printer = pprint.PrettyPrinter(width=78, compact=True)

                print('\n~ Trace ~')
                for item in state._call.trace:
                    print('\n<{link}> took {duration:.4f}s'.format(**item))
                    print('--')
                    printer.pprint(item['state'])
                print('\n~ fin ~\n')

            return r

        finally:
            if not (self._final is None):
                self._final(state)

    def __repr__(self):
        return 'Chain: {0}'.format(json.dumps(self.links, indent='    '))

    @property
    def links(self):
        """Return (a copy) of links in the chain"""
        return deepcopy(self._links)

    @links.setter
    def links(self, value):
        """Set the links in the chain (restructure)"""
        self._links = deepcopy(value)

    def copy(self):
        """Return a copy of the chain"""
        chain = self.__class__(self.links)
        chain._funcs = dict(self._funcs)
        return chain

    def get_final(self):
        """Get the final function for the chain"""
        return self._final

    def get_link(self, name):
        """Return the function associated with the named link"""
        return self._funcs.get(name)

    def insert_link(self, target_link, new_link, after=False):
        """
        Insert a new link in the chain before/after every instance of the target
        link. By default new links are inserted before the target link, to
        insert the new link after the target link pass `after` as True to the
        method.
        """
        parents = [self._links]
        indexes = [0]

        while True:
            index = indexes[-1]
            links = parents[-1]

            # Check if we're at the end of the current link list...
            if index >= len(links):
                # ...we are so pop this link list of the stack
                parents.pop()

                # Check if this was the last link list...
                if len(parents) == 0:
                    # ...it is so we're done
                    break
                else:
                    # ...it isn't so pop the index and continue with the
                    # previous link list.
                    indexes.pop()
                    continue

            # Get the current link
            link = parents[-1][index]
            indexes[-1] += 1

            # Check if the link is a intersection...
            if isinstance(link, list):
                # ...it is so add each fork to the stack
                parents += [link[1], link[0]]
                indexes += [0, 0]
                continue
            else:
                # ...it isn't so check if we need to add the new link after this
                # link.
                if link == target_link:
                    # Add the new link
                    if after:
                        parents[-1].insert(indexes[-1], new_link)
                    else:
                        parents[-1].insert(index, new_link)
                    indexes[-1] += 1

    def remove_link(self, target_link):
        """Remove all instances of a link from the chain"""
        parents = [self._links]
        indexes = [0]

        while True:
            index = indexes[-1]
            links = parents[-1]

            # Check if we're at the end of the current link list...
            if index >= len(links):
                # ...we are so pop this link list of the stack
                parents.pop()

                # Check if this was the last link list...
                if len(parents) == 0:
                    # ...it is so we're done
                    break
                else:
                    # ...it isn't so pop the index and continue with the
                    # previous link list.
                    indexes.pop()
                    continue

            # Get the current link
            link = parents[-1][index]
            indexes[-1] += 1

            # Check if the link is a intersection...
            if isinstance(link, list):
                # ...it is so add each fork to the stack
                parents += [link[1], link[0]]
                indexes += [0, 0]
                continue
            else:
                # ...it isn't so check if we need to remove the link.
                if link == target_link:
                    # Remove the link
                    parents[-1].remove(target_link)
                    indexes[-1] -= 1

    def set_final(self, func):
        """
        Set a function that is guaranteed to be called once the chain has
        executed (or failed to), think try...finally.
        """
        self._final = func

    def set_link(self, func, name=None):
        """
        Sets a function as a link. Works exactly like the `link()` decorator,
        e.g:

            def foo(state):
                ...

            chains.set_link(foo)

            # Set for the link named 'bar'
            def foo(state):
                ...

            chains.set_link(foo, name='bar')
        """
        self._funcs[name or func.__name__] = func

    def super(self, state):
        """
        Given an active state object, call the current link against this
        chain.

        It's often useful when overwriting a link to call the the original
        link, similar to calling super when overriding a method in a class
        you inherit from.

        This method is basically a shortcut for:

            my_chain.get_link('decorate')(state)

        Which becomes:

            my_chain.super(state)

        Active state objects contains a `_call` state with a list of links
        called so far (in `route`) and so `super` simply peeks at the last
        link in the route, finds the associated link in this chain and calls
        it with the given state.
        """

        assert '_call' in state, 'Not a valid state object'

        link = state._call['route'][-1]
        func = self.get_link(link)

        assert func, 'No corresponding link `{}` in chain'.format(link)

        return func(state)

    def update(self, links):
        """Update the chain with a new structured list of links"""
        self._links = links

    # Decorators

    def link(self, *args, **kwargs):
        """
        A decorator that will set the decorated function as a link, if no `name`
        keyword argument is specified then the function's name will be used to
        determine which link the function should be mapped to, e.g:

            @chain.link
            def foo(state):
                ...

            # Set for the link named 'bar'
            @chain.link(name='bar')
            def foo(state):
                ...
        """

        def wrap(func):
            self.set_link(func, **kwargs)
            return func

        if len(args) == 1 and callable(args[0]):
            return wrap(args[0])
        else:
            return wrap

    def final(self, func):
        """A decorator that will set the final function for the chain"""
        self.set_final(func)
        return func

    # Private methods

    def _call(self, _trace, state):
        """Call the chain of links"""

        # Call each link in the chain
        fork = None
        route = self.links
        while len(route):
            # Get the next link
            link = route.pop(0)

            # Check if the link is a fork
            if isinstance(link, list):

                assert len(link) == 2, \
                        'Intersections must have exactly 2 routes: {0}'\
                        .format(link)

                assert type(fork) == bool, \
                        'No fork determine at intesection: {0}'.format(link)

                # Fork to the new route
                route = link[0 if fork else 1]

                # Reset fork
                fork = None

                # Continue to the first link in the new route
                continue

            # Select the function for the link
            func = self._funcs.get(link)

            assert func is not None, 'Chain is incomplete: {0}'.format(link)

            # Track the link against the scope
            state._call.route.append(link)

            if _trace:
                # If the call is being traced record the point at which we
                # called this link.
                start = time.time()

            # Call the link function
            result = func(state)

            if _trace:
                # If the call is being traced record the link, it's call
                # duration and the state after the call.

                state_copy = state.copy()
                del state_copy['_call']

                state._call.trace.append({
                    'link': link,
                    'duration': time.time() - start,
                    'state': state_copy
                })

            # Determine the next step...

            if result is None:
                # ...no result returned so continue to the next link
                continue

            elif type(result) is bool:
                # ...a boolean was returned so the next link will be an
                # intersection and the fork taken will be determine by the
                # result.
                fork = result
                continue

            # ...something else was returned so we leave break from the chain
            # and return the result.
            return result


class State(dict):
    """
    An dictionary-like object that supports dot notation syntax. State objects
    are used to pass state between links within a chain.
    """

    def __getattr__(self, attr):
        if attr.startswith('__'):
            raise AttributeError
        return self.get(attr, None)

    __setattr__ = dict.__setitem__

    __delattr__ = dict.__delitem__
